﻿//Version 1.1.3.0
using System;
using System.Data;
using System.Linq;
using System.Xml;
using System.IO;
using System.Net;
using System.Globalization;
using System.Reflection;
using System.Collections.Generic;
using System.Security;
using System.Threading;
using System.Threading.Tasks;

using Genetec.Sdk;
using Genetec.Sdk.Entities;
using Genetec.Sdk.Entities.Video;
using Genetec.Sdk.Entities.Video.HardwareSpecific.Axis;
using Genetec.Sdk.Entities.Video.VisualTracking;
using Genetec.Sdk.Enumerations;
using Genetec.Sdk.Queries;
using Genetec.Sdk.Scripting;
using Genetec.Sdk.Workflows.UnitManager;
using Genetec.Sdk.Diagnostics.Logging;
using Genetec.Sdk.Diagnostics.Logging.Core;


public static class MacroVariables
{
	//Please enter the IP Adress of the Panomera main module, compile, and execute the macro.
	//without leading zeros
	//example: "127.0.0.1" and not "127.000.000.001"

	public static string ipAddressMainModule = "10.128.80.60";

	//Thickness of the border of VisualTracking zones drawn on the main module
	public static int strokeThickness = 2;

	// "colors" contains pairs of colors for fill and stroke: ( Fillcolor, Strokecolor )
	// "colors" must contain at least one pair.
	// The macro loops over the pairs in "colors".
	//   It is possible to add more pairs.
	//   It is possible to define only one pair so all VisualTracking zones get the same color:
	//      public static (string, string)[] colors =
	//      {
	//          ( "#01000000", "#01000000" )
	//			//( "#011CBBFF", "#FF1CBBFF" )
	//      };
	// Colors are given by in the format "#AARRGGBB"
	//   components are given in 2 digit hexadecimal from 00 to FF
	//     AA: Alpha defined the transparency: 00=transparent, FF=opaque
	//         fully transparent zones are not clickable. It is recommended to set transparency=01
	//     RR: red component
	//     GG: green component
	//     BB: blue component
	public static (string, string) [] colors = 
	{
	    ( "#01000000", "#FF1CBBFF" )
	/*	( "#40FF0000", "#FFFF0000" ),
		( "#4000FF00", "#FF00FF00" ),
		( "#400000FF", "#FF0000FF" ),
		( "#40FFFF00", "#FFFFFF00" ),
		( "#40FF00FF", "#FFFF00FF" ),
		( "#4000FFFF", "#FF00FFFF" ),
		( "#40800000", "#FF800000" )
	*/
	};
	//opacity of the shape in %: 0=transparent, 100=opaque
	public static int zoneOpacity = 60; 

	//if true: configure VisualTracking zones from Panomera sub modules back to the Panomera main module.
	//if false: zones on sub modules are not changed
	public static bool addVTZoneFromSub2Main = true; // = false;
	public static (string, string) colorReverse = ("#01000000", "#01000000");
	public static int strokeThicknessReverse = 0;
	public static int zoneOpacityReverse = 60; 
}

public class Geometry
{
	public int width { get; set; }
	public int height { get; set; }
	public int clientRotation { get; set; }
	public List<Module> modules { get; set; }
}

public class Module
{
	public int index { get; set; }
	public string ipaddr { get; set; }
	public int left { get; set; }
	public int top { get; set; }
	public int width { get; set; }
	public int height { get; set; }
	public int clientRotation { get; set; }
}

public sealed class MyMacro : UserMacro
{
	private string _login = "admin";
	public string Login
	{
		get { return _login; }
		set
		{
			if (string.IsNullOrEmpty(value) == false)
				_login = value;
		}
	}

	private string _password = "3";
	public string Password
	{
		get { return _password; }
		set
		{
			if (string.IsNullOrEmpty(value) == false)
				_password = value;
		}
	}

	private int _postNr = 80;
	public int Port
	{
		get { return _postNr; }
		set
		{
			if (value > 0)
				_postNr = value;
		}
	}

	public bool IsDebugMode { get; set; }

	/// <summary>
	/// Entry point of the macro. Provide an implementation of this method.
	/// </summary>
	public override void Execute()
	{
		try
		{
			UpdateEntities();

			LogEntry(LogSeverity.Information, "Macro Started");

			IList<VideoUnit> videoUnits = Sdk.GetEntities<VideoUnit>(EntityType.VideoUnit);

			bool unitFound = false;

			LogEntry(LogSeverity.Information, $"Found video units: {videoUnits.Count}");
			foreach (VideoUnit videoUnit in videoUnits)
			{
				if (string.Equals(videoUnit.IPAddress, MacroVariables.ipAddressMainModule))
				{
					LogEntry(LogSeverity.Information, $"Video unit: Name: {videoUnit.Name} ({videoUnit.IPAddress})");
					ConfigureVirtualZone(videoUnit);
					unitFound = true;
					break;
				}
			}
			if(!unitFound)
			{
				LogEntry(LogSeverity.Information, $"Unit with IP: {MacroVariables.ipAddressMainModule} not found!");
			}
			LogEntry(LogSeverity.Information, "Macro Stopped");
		}
		catch (Exception ex)
		{
			string error = ex.Message;
			LogEntry(LogSeverity.Error, "Exception : " + error, ex);
		}
	}

	protected void ConfigureVirtualZone(VideoUnit videoUnit)
	{
		Camera lCamera = Sdk.GetEntity(videoUnit.Cameras.First()) as Camera;
		if (null == lCamera)
		{
			LogEntry(LogSeverity.Error, $"VideoUnit ({videoUnit.IPAddress}) has no camera.");
			return;
		}

		Geometry geometry = GetAvailableModules(videoUnit.IPAddress);
		if (null == geometry || null == geometry.modules || geometry.modules.Count() <= 1)
		{
			LogEntry(LogSeverity.Error, $"VideoUnit ({videoUnit.IPAddress}) can't be preconfigured by this macro.");
			return;
		}

		VisualTrackingConfiguration lConfig = lCamera.GetVisualTrackingConfiguration();
		lConfig.ClearZones();

		int colorCount = MacroVariables.colors.GetLength(0);
		foreach (Module module in geometry.modules)
		{
			LogEntry(LogSeverity.Information, $"Adding VisualTrackingZone : {module.index} - IP: {module.ipaddr} (X: {module.left}, Y: {module.top}, Width: {module.width}, Height: {module.height})");

			if (0 == module.index)
			{
				LogEntry(LogSeverity.Information, $"{module.ipaddr} is a main sensor.");
				continue;
			}

			VisualTrackingZone zone = new VisualTrackingZone();

			zone.RelativeXPosition = VisualTrackingZone.ConvertToRelativePosition(module.left, module.width, geometry.width);
			zone.RelativeYPosition = VisualTrackingZone.ConvertToRelativePosition(module.top, module.height, geometry.height);
			zone.StrokeThickness = MacroVariables.strokeThickness;
			zone.RelativeWidth = ((double)module.width / (double)geometry.width) * 100.0;
			zone.RelativeHeight = ((double)module.height / (double)geometry.height) * 100.0;
			int colorIndex = (module.index - 1) % colorCount;
			zone.FillColor = MacroVariables.colors[colorIndex].Item1;
			zone.StrokeColor = MacroVariables.colors[colorIndex].Item2;
			zone.Opacity = MacroVariables.zoneOpacity;
			try
			{
				zone.AddEntity(GetCameraGuid(videoUnit.IPAddress, module.ipaddr, videoUnit.ArchiverRole));
				lConfig.AddZone(zone);
			}
			catch (Exception ex)
			{
				LogEntry(LogSeverity.Error, $"Can't get camera. Error: {ex.Message}", ex);
				LogEntry(LogSeverity.Information, $"Assotiation with a camera will be skipped...");
			}

			if(MacroVariables.addVTZoneFromSub2Main)
			{
				VideoUnit subUnit = GetVideoUnitByIP(module.ipaddr);
				if (null != subUnit)
				{
					VisualTrackingZone zone2 = new VisualTrackingZone();

					zone2.RelativeXPosition = 50;
					zone2.RelativeYPosition = 50;
					zone2.StrokeThickness = MacroVariables.strokeThicknessReverse;
					zone2.RelativeWidth = 100;
					zone2.RelativeHeight = 100;
					zone2.FillColor = MacroVariables.colorReverse.Item1;
					zone2.StrokeColor = MacroVariables.colorReverse.Item2;
					zone2.Opacity = MacroVariables.zoneOpacityReverse;
					try
					{
						Camera subCamera = Sdk.GetEntity(subUnit.Cameras.First()) as Camera;
						if (subCamera == null)
						{
							LogEntry(LogSeverity.Error, $"VideoUnit sub module ({subUnit.IPAddress}) has no camera.");
						}
						else
						{
							VisualTrackingConfiguration subVTConfig = subCamera.GetVisualTrackingConfiguration();
							subVTConfig.ClearZones();

							zone2.AddEntity(GetCameraGuid(module.ipaddr, videoUnit.IPAddress,subUnit.ArchiverRole));
							subVTConfig.AddZone(zone2);
							subCamera.SetVisualTrackingConfiguration(subVTConfig);
							LogEntry(LogSeverity.Information, $"Added VisualTrackingZone from sub module {subUnit.IPAddress} to main module {videoUnit.IPAddress}");
						}
					}
					catch (Exception e)
					{
						LogEntry(LogSeverity.Error, $"Can't get camera. Error: {e.Message}", e);
						LogEntry(LogSeverity.Information, $"skipping link back from sub module {subUnit.IPAddress} to main module {videoUnit.IPAddress}");
					}
				}
			}
		}

		try
		{
			// Save VisualTracking Zones to the camera.
			lCamera.SetVisualTrackingConfiguration(lConfig);
		}
		catch (Exception e)
		{
			LogEntry(LogSeverity.Error, $"Can't save VisualTrackingConfiguration: {e.Message}", e);
			if (null != e.InnerException)
				LogEntry(LogSeverity.Error, $"InnerException: e.InnerException.Message", e.InnerException);
		}
	}

	protected Guid GetCameraGuid(string masterIP, string cameraIP, Role archiverRole)
	{
		VideoUnit videoUnit = GetVideoUnitByIP(cameraIP);
		if (null == videoUnit)
		{
			videoUnit = CreateVideoUnit(masterIP, cameraIP, archiverRole);
		}

		if (null != videoUnit)
		{
			return videoUnit.Cameras.First();
		}

		throw new Exception($"Can't find any camera with IP {cameraIP}");
	}

	private VideoUnit GetVideoUnitByIP(string ipAddress)
	{
		foreach (VideoUnit videoUnit in Sdk.GetEntities(EntityType.VideoUnit))
		{
			if (string.Equals(videoUnit.IPAddress, ipAddress))
			{
				LogEntry(LogSeverity.Information, $"Found a video unit. IP {videoUnit.IPAddress}");
				return videoUnit;
			}
		}

		LogEntry(LogSeverity.Warning, $"VideoUnit with IP {ipAddress} does not exist.");

		return null;
	}


	private AutoResetEvent _syncEventSuccess = new AutoResetEvent(false);
	private AutoResetEvent _syncEventFail = new AutoResetEvent(false);

	private void UnitEnrolledEventHandler(object o, Genetec.Sdk.EventsArgs.UnitEnrolledEventArgs e)
	{
		if (EnrollmentResult.Connecting == e.EnrollmentResult)
		{
			// started connection
			LogEntry(LogSeverity.Debug, $"VideoUnit {e.Address}:{e.Port} trying to connect...");
		}
		else
		{
			if (EnrollmentResult.Added != e.EnrollmentResult)
			{
				LogEntry(LogSeverity.Debug, $"VideoUnit {e.Address}:{e.Port} was not added. Error: {e.EnrollmentResult}");
				_syncEventFail.Set();
			}
			else
			{
				LogEntry(LogSeverity.Debug, $"VideoUnit {e.Address}:{e.Port} was successfully added.");
				_syncEventSuccess.Set();
			}
		}
	}

	private VideoUnit CreateVideoUnit(string masterIP, string ipAddress, Role archiverRole)
	{
		VideoUnitProductInfo vuProductInfo = Sdk.VideoUnitManager.FindProductsByManufacturer("ONVIF")?.SingleOrDefault(x => x.ProductType.Equals("All")); ;
		if (vuProductInfo != null)
		{
			try
			{
				AddVideoUnitInfo videoUnitInfo = new AddVideoUnitInfo(vuProductInfo, new IPEndPoint(IPAddress.Parse(ipAddress), Port), false);
				videoUnitInfo.UserName = Login;
				videoUnitInfo.Password = new SecureString();
				foreach (char ch in Password)
					videoUnitInfo.Password.AppendChar(ch);

				Sdk.VideoUnitManager.EnrollmentStatusChanged += UnitEnrolledEventHandler;

				// OK, here is a trick.
				// AddVideoUnit - triggers an asynchron routine.
				// AddUnitResponse notifies, that it was possible to start adding of a new video unit.
				// But the result of this action will be notified later by EnrollmentStatusChanged-event.
				// My goal here - create CreateVideoUnit synchron:
				// 1) Call AddVideoUnit
				AddUnitResponse response = Sdk.VideoUnitManager.AddVideoUnit(videoUnitInfo, archiverRole.Guid).Result;
				if (Error.None != response.Error)
				{
					throw new Exception($"Can't add video unit (IP: {ipAddress}).\r\n\t- Error: {response.Error.ToString()}\r\n\t- MissingInformation: {response.MissingInformation}");
				}

				// 2) Start new Task. This Task waits until SDK notifies us via EnrollmentStatusChanged-event, that the video unis was sucussfully added (or failed to add the video unit)
				bool isSuccess = Task.Run<bool>(() =>
				{
					int eventIndex = WaitHandle.WaitAny(new WaitHandle[] { _syncEventSuccess, _syncEventFail });

					return 0 == eventIndex;
				}).Result;

				// 3) Adding of the video unit is ...
				VideoUnit videoUnit = null;
				if (true == isSuccess)
				{
					// 3.1) ... sucessfully finished.
					LogEntry(LogSeverity.Information, $"*** VideoUnit ({ipAddress}) was added.");

					// 4) Update entities
					UpdateEntities();

					// 5) Rename new vidoe unit
					videoUnit = GetVideoUnitByIP(ipAddress);
					videoUnit.Name = $"{masterIP} - SUB ({ipAddress})";
				}
				else
				{
					// 3.2) ... failed.
					LogEntry(LogSeverity.Error, $"Can't add video unit (IP: {ipAddress}).\r\n\t- Check if ONVIF for the camera is enabled.");
				}
				return videoUnit;
			}
			finally
			{
				Sdk.VideoUnitManager.EnrollmentStatusChanged -= UnitEnrolledEventHandler;
			}
		}

		throw new Exception("Can't get ONVIF product info");
	}

	/// <summary>
	/// Called when the macro needs to clean up. 
	/// </summary>    
	protected override void CleanUp()
	{
		// Release objects created by the Execute method, unhook events, and so on.
	}

	private void UpdateEntities()
	{
		EntityConfigurationQuery query = Sdk.ReportManager.CreateReportQuery(ReportType.EntityConfiguration) as EntityConfigurationQuery;
        query.EntityTypes.Add(EntityType.VideoUnit);
		query.DownloadAllRelatedData = true;
		QueryCompletedEventArgs result = query.Query();
	}

	private void LogEntry(LogSeverity logSeverity, string logEntry)
	{
		LogEntry(logSeverity, logEntry, null);
	}
	private void LogEntry(LogSeverity logSeverity, string logEntry, Exception ex)
	{
		if (false == IsDebugMode)
			MacroLogger.Trace(logSeverity, new TraceEntry(logEntry, string.Empty, ex));
		else
			MacroLogger.Trace((LogSeverity)((int)logSeverity * 2), new TraceEntry(logEntry, string.Empty, ex));
	}

	private Geometry GetAvailableModules(string hostName)
	{
		Geometry geometry = null;

		try
		{
			string webServiceUrl = $"http://{hostName}/geometry";
			WebRequest request = WebRequest.Create(webServiceUrl);
			request.Credentials = new NetworkCredential(Login, Password);
			using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
			{
				if (HttpStatusCode.OK == response.StatusCode && response.ContentType.ToLower().Contains("json"))
				{
					using (System.IO.Stream dataStream = response.GetResponseStream())
					{
						// Load required assembly
						Assembly asmDataContractJson = Assembly.Load("System.Runtime.Serialization.Json, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a");

						// Instantiate DataContractJsonSerializer
						Type typeDataContractJsonSerializer = asmDataContractJson.GetType("System.Runtime.Serialization.Json.DataContractJsonSerializer");
						object objDataContractJsonSerializer = Activator.CreateInstance(typeDataContractJsonSerializer, new object[] { typeof(Geometry) });

						// Call ReadObject
						MethodInfo methodReadObject = typeDataContractJsonSerializer.GetMethod("ReadObject", new Type[] { typeof(System.IO.Stream) });

						geometry = methodReadObject.Invoke(objDataContractJsonSerializer, new object[] { dataStream }) as Geometry;

						// validate recieved data
						if (0 >= geometry.height || 0 >= geometry.width)
						{
							LogEntry(LogSeverity.Error, $"Invalid geometry. Please ensure, that WebService {webServiceUrl} works correctly.");
							return null;
						}
					}
				}
			}
		}
		catch (Exception ex)
		{
			LogEntry(LogSeverity.Error, $"Can't get available modules. Error: {ex.Message}", ex);
		}
		return geometry;
	}

}
